"""
This module contains functions to scrape the Json toi/shifts for any given game
"""

import pandas as pd
import json
import hockey_scraper.shared as shared


def get_shifts(game_id):
    """
    Given a game_id it returns the raw json
    Ex: http://www.nhl.com/stats/rest/shiftcharts?cayenneExp=gameId=2010020001
    
    :param game_id: the game
    
    :return: json or None
    """
    page_info = {
        "url": 'http://www.nhl.com/stats/rest/shiftcharts?cayenneExp=gameId={}'.format(game_id),
        "name": str(game_id),
        "type": "json_shifts",
        "season": str(game_id)[:4],
    }

    response = shared.get_file(page_info)

    # Return empty dict if can't get page
    if not response:
        return {}
    else:
        return json.loads(response)


def fix_team_tricode(tricode):
    """
    Some of the tricodes are different than how I want them
    
    :param tricode: 3 letter team name - ex: NYR
    
    :return: fixed tricode
    """
    fixed_tricodes = {
        'TBL':  'T.B',
        'LAK': 'L.A',
        'NJD': 'N.J',
        'SJS': 'S.J'
    }

    if tricode.upper() in list(fixed_tricodes.keys()):
        return fixed_tricodes[tricode.upper()]
    else:
        return tricode


def parse_shift(shift):
    """
    Parse shift for json
    
    :param shift: json for shift
    
    :return: dict with shift info
    """
    shift_dict = dict()

    name = shared.fix_name(' '.join([shift['firstName'].strip(' ').upper(), shift['lastName'].strip(' ').upper()]))
    shift_dict['Player'] = name
    shift_dict['Player_Id'] = shift['playerId']
    shift_dict['Period'] = shift['period']
    shift_dict['Team'] = fix_team_tricode(shift['teamAbbrev'])

    # At the end of the json they list when all the goal events happened. They are the only one's which have their
    # eventDescription be not null
    if shift['eventDescription'] is None:
        shift_dict['Start'] = shared.convert_to_seconds(shift['startTime'])
        shift_dict['End'] = shared.convert_to_seconds(shift['endTime'])
        shift_dict['Duration'] = shared.convert_to_seconds(shift['duration'])
    else:
        shift_dict = dict()

    return shift_dict


def parse_json(shift_json, game_id):
    """
    Parse the json
    
    :param shift_json: raw json
    :param game_id: if of game
    
    :return: DataFrame with info
    """
    columns = ['Game_Id', 'Period', 'Team', 'Player', 'Player_Id', 'Start', 'End', 'Duration']

    shifts = [parse_shift(shift) for shift in shift_json['data']]        # Go through the shifts
    shifts = [shift for shift in shifts if shift != {}]                  # Get rid of null shifts (which happen at end)

    df = pd.DataFrame(shifts, columns=columns)
    df['Game_Id'] = str(game_id)[5:]
    df = df.sort_values(by=['Period', 'Start'], ascending=[True, True])  # Sort by period by time
    df = df.reset_index(drop=True)

    return df


def scrape_game(game_id):
    """
    Scrape the game. 
    
    :param game_id: game
    
    :return: DataFrame with info for the game
    """
    shifts_json = get_shifts(game_id)

    if not shifts_json:
        shared.print_warning("Json shifts for game {} is either not there or can't be obtained".format(game_id))
        return None

    try:
        game_df = parse_json(shifts_json, game_id)
    except Exception as e:
        shared.print_warning('Error parsing Json shifts for game {} {}'.format(game_id, e))
        return None

    return game_df if not game_df.empty else None



from hockey_scraper.scrape_functions import scrape_games, scrape_date_range, scrape_seasons
from hockey_scraper.live_scrape import ScrapeLiveGames, LiveGame
from hockey_scraper.nwhl import scrape_functions as nwhl



"""Saves the scraped docs so you don't have to re-scrape them every time you want to parse the docs. 

This only works if you specified a valid directory in hockey_scraper.shared.docs_dir. If not nothing will be saved or
be retrieved (if it exists but you didn't set it)

\**** Don't mess with this unless you know what you're doing \****
"""
import os


def create_file_path(file_info):
    """
    Creates the file path for a given file
    
    :param file_info: Dictionary containing the info on the file. Includes the name, season, file type, and the dir
                      we want to deposit any data in.
    
    :return: path 
    """
    # Shitty fix for when you already have it saved but don't have nwhl folders
    if 'nwhl' in file_info['type']:
        if not os.path.isdir(os.path.join(file_info['dir'], '/'.join(['docs', str(file_info['season']), file_info['type']]))):
            os.mkdir(os.path.join(file_info['dir'], '/'.join(['docs', str(file_info['season']), file_info['type']])))

    return os.path.join(file_info['dir'],
                        '/'.join(['docs', str(file_info['season']), file_info['type'], file_info['name'] + ".txt"]))


def create_season_dirs(season):
    """
    Creates the infrastructure to hold all the scraped docs for a season
    
    :param season: given season
                        
    :return: None
    """
    os.chdir(os.path.join(os.getcwd(), "docs"))

    # Folder for season
    os.mkdir(str(season))
    os.chdir(str(season))

    # Create all sub-folders
    os.mkdir("html_pbp")
    os.mkdir("json_pbp")
    os.mkdir("espn_pbp")
    os.mkdir("html_shifts_home")
    os.mkdir("html_shifts_away")
    os.mkdir("json_shifts")
    os.mkdir("html_roster")
    os.mkdir("json_schedule")
    os.mkdir("espn_scoreboard")

    # Move back to the previous directory
    os.chdir('..')


def check_file_exists(file_info):
    """
    Checks if the file exists. Also check if structure for holding scraped file exists to. If not, it creates it. 

    :param file_info: Dictionary containing the info on the file. Includes the name, season, file type, and the dir
                      we want to deposit any data in.

    :return: Boolean - True if it exists
    """
    # Create the docs subdir if it doesn't exist
    if not os.path.isdir(os.path.join(file_info['dir'], 'docs')):
        os.mkdir("docs")

    # Check if the folder for the season for the given game was created yet...if not create it
    if not os.path.isdir(os.path.join(file_info['dir'], '/'.join(['docs', str(file_info['season'])]))):
        create_season_dirs(file_info['season'])

    return os.path.isfile(create_file_path(file_info))


def get_page(file_info):
    """
    Get the file so we don't need to re-scrape 

    :param file_info: Dictionary containing the info on the file. Includes the name, season, file type, and the dir
                      we want to deposit any data in.

    :return: Response or None
    """
    with open(create_file_path(file_info), 'r') as my_file:
        return my_file.read().replace('\n', '')


def save_page(page, file_info, docs_dir):
    """
    Save the page we just scraped.
    
    Note: It'll only get saved if the directory already exists!!!!!!. I'm not dealing with any fuck ups. That would 
    involve checking if it's even a valid path and creating it. Make sure you get it right. 

    :param page: File scraped
    :param file_info: Dictionary containing the info on the file. Includes the name, season, file type, and the dir
                      we want to deposit any data in.
    :param docs_dir: Directory specified - either none or a valid dir

    :return: None
    """
    if docs_dir is not None and page is not None and page != '':
        with open(create_file_path(file_info), 'w') as file:
            file.write(page)



"""
This module contains code to scrape coordinates for games off of espn for any given game
"""

import pandas as pd
from bs4 import BeautifulSoup
import xml.etree.ElementTree as etree
import re
import hockey_scraper.shared as shared


def event_type(play_description):
    """
    Returns the event type (ex: a SHOT or a GOAL...etc) given the event description 
    
    :param play_description: description of play
    
    :return: event
    """
    events = {'GOAL SCORED': 'GOAL', 'SHOT ON GOAL': 'SHOT', 'SHOT MISSED': 'MISS', 'SHOT BLOCKED': 'BLOCK',
              'PENALTY': 'PENL', 'FACEOFF': 'FAC', 'HIT': 'HIT', 'TAKEAWAY': 'TAKE', 'GIVEAWAY': 'GIVE'}

    event = [events[e] for e in events.keys() if e in play_description]
    return event[0] if event else None


def get_game_ids(response):
    """
    Get game_ids for date from doc
    
    :param response: doc
    
    :return: list of game_ids
    """
    soup = BeautifulSoup(response, 'lxml')

    divs = soup.findAll('div', {'class': "game-header"})
    regex = re.compile(r'id="(\d+)')
    game_ids = [regex.findall(str(div))[0] for div in divs]

    return game_ids


def get_teams(response):
    """
    Extract Teams for date from doc
    
    :param response: doc
    
    :return: list of teams    
    """
    soup = BeautifulSoup(response, 'lxml')

    td = soup.findAll('td', {'class': "team"})
    teams = [shared.get_team(t.get_text().upper()) for t in td if t.get_text() != '']

    # Make a list of both teams for each game
    games = [teams[i:i + 2] for i in range(0, len(teams), 2)]

    return games


def get_espn_date(date):
    """
    Get the page that contains all the games for that day
    
    :param date: YYYY-MM-DD
    
    :return: response 
    """
    page_info = {
        "url": 'http://www.espn.com/nhl/scoreboard?date={}'.format(date.replace('-', '')),
        "name": date,
        "type": "espn_scoreboard",
        "season": shared.get_season(date),
    }
    response = shared.get_file(page_info)

    # If can't get or not there throw an exception
    if not response:
        raise Exception
    else:
        return response


def get_espn_game_id(date, home_team, away_team):
    """
    Scrapes the day's schedule and gets the id for the given game
    Ex: http://www.espn.com/nhl/scoreboard?date=20161024
    
    :param date: format-> YearMonthDay-> 20161024
    :param home_team: home team
    :param away_team: away team
    
    :return: 9 digit game id as a string
    """
    response = get_espn_date(date)
    game_ids = get_game_ids(response)
    games = get_teams(response)

    # Get the game id with the right team for it
    for i in range(len(games)):
        if home_team in games[i] or away_team in games[i]:
            return game_ids[i]


def get_espn_game(date, home_team, away_team, game_id=None):
    """
    Gets the ESPN pbp feed 
    Ex: http://www.espn.com/nhl/gamecast/data/masterFeed?lang=en&isAll=true&gameId=400885300
    
    :param date: date of the game
    :param home_team: home team
    :param away_team: away team
    :param game_id: Game id of we already have it - for live scraping. None if not there
    
    :return: raw xml
    """
    # Get if not provided
    if not game_id:
        game_id = get_espn_game_id(date, home_team.upper(), away_team.upper())

    file_info = {
        "url": 'http://www.espn.com/nhl/gamecast/data/masterFeed?lang=en&isAll=true&gameId={}'.format(game_id),
        "name": game_id,
        "type": "espn_pbp",
        "season": shared.get_season(date),
    }
    response = shared.get_file(file_info)

    if response is None:
        raise Exception

    return response


def parse_event(event):
    """
    Parse each event. In the string each field is separated by a '~'. 
    Relevant for here: The first two are the x and y coordinates. And the 4th and 5th are the time elapsed and period.
    
    :param event: string with info
    
    :return: return dict with relevant info
    """
    info = dict()
    fields = event.split('~')

    # Shootouts screw everything up so don't bother...coordinates don't matter there either way
    if fields[4] == '5':
        return None

    info['xC'] = float(fields[0])
    info['yC'] = float(fields[1])
    info['time_elapsed'] = shared.convert_to_seconds(fields[3])
    info['period'] = fields[4]
    info['event'] = event_type(fields[8].upper())

    return info


def parse_espn(espn_xml):
    """
    Parse feed 
    
    :param espn_xml: raw xml of feed
    
    :return: DataFrame with info
    """
    columns = ['period', 'time_elapsed', 'event', 'xC', 'yC']

    # Occasionally we get malformed XML because of the presence of \x13 characters
    # Let's just replace them with dashes
    espn_xml = espn_xml.replace(u'\x13', '-')

    try:
        tree = etree.fromstring(espn_xml)
    except etree.ParseError:
        shared.print_warning("Espn pbp isn't valid xml, therefore coordinates can't be obtained for this game")
        return pd.DataFrame([], columns=columns)

    events = tree[1]
    plays = [parse_event(event.text) for event in events]
    plays = [play for play in plays if play is not None]    # Get rid of plays that are None

    return pd.DataFrame(plays, columns=columns)


def scrape_game(date, home_team, away_team, game_id=None):
    """
    Scrape the game
    
    :param date: ex: 2016-20-24
    :param home_team: tricode
    :param away_team: tricode
    :param game_id: Only provided for live games.
    
    :return: DataFrame with info 
    """
    try:
        shared.print_warning('Using espn for pbp')
        espn_xml = get_espn_game(date, home_team, away_team, game_id)
    except Exception as e:
        shared.print_warning("Espn pbp for game {a} {b} {c} is either not there or can't be obtained {d}".format(a=date,
                                                                                                    b=home_team,
                                                                                                    c=away_team, d=e))
        return None

    try:
        espn_df = parse_espn(espn_xml)
    except Exception as e:
        shared.print_warning("Error parsing Espn pbp for game {a} {b} {c} {d}".format(a=date, b=home_team, c=away_team, d=e))
        return None

    return espn_df



"""
Module to scrape live game info
"""
import pandas as pd
import datetime
import warnings
import time

import hockey_scraper.json_schedule as json_schedule
import hockey_scraper.espn_pbp as espn_pbp
import hockey_scraper.playing_roster as playing_roster
import hockey_scraper.game_scraper as game_scraper
import hockey_scraper.json_pbp as json_pbp
import hockey_scraper.shared as shared


def set_docs_dir(user_dir):
    """
    Set the docs directory
    
    :param user_dir: User specified directory for storing saves scraped files
    
    :return: None
    """
    # We always want to rescrape since the files are being updated constantly
    shared.if_rescrape(True)

    shared.add_dir(user_dir)


def check_date_format(date):
    """
    Verify the date format. If wrong raises a HaltException

    :param date: User supplied date

    :return: None
    """
    try:
        time.strptime(date, "%Y-%m-%d")
    except ValueError:
        raise shared.HaltException("Error: Incorrect format given for dates. They must be given like 'yyyy-mm-dd' "
                                   "(ex: '2016-10-01').")


class LiveGame:
    """ 
    This is a class holds all the information for a given game
      
    :param int game_id: The NHL game id (ex: 2018020001)
    :param datetime start_time: The UTC time of when the game begins
    :param str home_team: Tricode for the home team (ex: NYR)
    :param str away_team: Tricode for the home team (ex: MTL)
    :param int espn_id: The ESPN game id for their feed
    :param str date: Date of the game (ex: 2018-10-30)
    :param bool if_scrape_shifts: Whether or not you want to scrape shifts
    :param str api_game_status: Current Status of the game - ["Final", "Live", "Intermission]
    :param str html_game_status: Current Status of the game - ["Final", "Live", "Intermission"]
    :param int intermission_time_remaining: Time remaining in the intermission. 0 if not in intermission
    :param dict players: Player info for both teams
    :param dict head_coaches: Head coaches for both teams
    :param DataFrame pbp_df: Holds most recent pbp data
    :param DataFrame shifts_df: Holds most recent shift data
    :param DataFrame prev_pbp_df: Holds the previous pbp data (for just in case)
    :param DataFrame prev_shifts_df: Holds the previous shift data (for just in case)
    """

    def __init__(self, game_id, start_time, home_team, away_team, status, espn_id, date, if_scrape_shifts):
        """ Constructor """
        # Given upon creation
        self.game_id = game_id
        self.start_time = start_time
        self.home_team = home_team
        self.away_team = away_team
        self.espn_id = espn_id
        self.date = date
        self.if_scrape_shifts = if_scrape_shifts

        # Html pbp is behind the json (json updates faster)
        self.api_game_status = status
        self.html_game_status = "Live"
        self.prev_api_game_status = status
        self.prev_html_game_status = "Live"
        self.intermission_time_remaining = 0

        # We know nothing to start off
        self.players = None
        self.head_coaches = None

        # Pbp and shift data - Will be filled in later
        # Also hold previous pair for checking for changes
        self.pbp_df = pd.DataFrame()
        self.shifts_df = pd.DataFrame()
        self.prev_pbp_df = pd.DataFrame()
        self.prev_shifts_df = pd.DataFrame()

        # Object creation message
        print("The LiveGame object for game {game_id} has been created. ".format(game_id=game_id))
        print("Game starts in {time} seconds.".format(time=self.time_until_game()))

    def scrape(self, force=False):
        """
        Scrape the given game. Check if currently ongoing or started
        
        :param bool force: Whether or not to force it to scrape even if it's over
        
        :return: None
        """
        # 1. force = False: If the game hasn't eclipsed the starting time or is over we don't scrape
        # 2. force = True: We always scrape
        if (self.time_until_game() == 0 and not self.is_game_over(prev=True)) or force:
            self.scrape_live_game(force=force)

    def scrape_live_game(self, force=False):
        """
        Scrape the live info for a given game
        
        :param force: Whether to scrape no matter what (used for intermission here)
        
        :return: None
        """
        game_json = json_pbp.get_pbp(str(self.game_id))

        # When don't have json...can't do anything without it
        if game_json is None:
            return

        # Shift Game Statuses b4 we do anything
        self.prev_api_game_status = self.api_game_status
        self.prev_html_game_status = self.html_game_status

        # Swap old pbp & shift DataFrames
        self.prev_pbp_df = self.pbp_df
        self.prev_shifts_df = self.shifts_df

        # If json is in intermission:
        # Update self.api_game_status, get minutes remaining in intermission, and check if html is intermission too.
        # If both feeds are in intermission we return, otherwise we wait for the html to catch up.
        # "Intermission" is my own game status so otherwise just take whatever is in the api
        if game_json['liveData']['linescore']['intermissionInfo']['inIntermission']:
            self.api_game_status = "Intermission"
            self.intermission_time_remaining = game_json['liveData']['linescore']['intermissionInfo']["intermissionTimeRemaining"]

            # If see the both says intermission and we do too, we can just safely return and not bother with scraping.
            # This will be false if the HTML hasn't updated yet to intermission
            # If force we scrape no matter what
            if self.is_intermission() and not force:
                return
        else:
            # Update API Status if NOT in intermission to whatever is there
            self.api_game_status = game_json["gameData"]["status"]["abstractGameState"]

        # Leave if b4 game started
        if game_json["gameData"]["status"]["abstractGameState"] in ["Preview"]:
            self.html_game_status = self.api_game_status = game_json["gameData"]["status"]["abstractGameState"]
            return

        # We get this the 1st time it scrapes the info (or when it's first available)
        # Don't bother with earlier as it may not be there or we may end up with an old version
        if not self.players:
            roster = playing_roster.scrape_roster(self.game_id)
            if roster is not None:
                self.players, _ = game_scraper.get_teams_and_players(game_json, roster, self.game_id)
                self.head_coaches = roster['head_coaches']
            else:
                return  # If we try and still can't get it we leave - Termination Reason #2

        # Don't bother with scraper warnings
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            # Scrape pbp - pay attention to each argument
            self.pbp_df, self.html_game_status = game_scraper.scrape_pbp_live(self.game_id, self.date,
                                                                              {"head_coaches": self.head_coaches},
                                                                              game_json, self.players,
                                                                              {"Home": self.home_team, "Away": self.away_team},
                                                                              espn_id=self.espn_id)
            # Get shifts if asked for
            if self.if_scrape_shifts:
                self.shifts_df = game_scraper.scrape_shifts(self.game_id, self.players, self.date)

    def is_ongoing(self):
        """
        Check if the game is currently being played. 
        
        The logic here is that we run into an issue with intermission and the end of game. If the game is just changed 
        to Final or Intermission the end user will assume the game isn't ongoing and will not update with the most
        recent events. They'll be delayed for intermission and won't place it at all for Final games. So we use the 
        previous event as a guide. If it's currently in intermission or Final - we check the previous status. If it's 
        the same the user already has the data. Otherwise we 'lie' and say the game is still ongoing.
        
        :return: Boolean
        """
        # The game is currently being played
        if self.time_until_game() == 0 and not self.is_game_over() and not self.is_intermission() and self.pbp_df.shape[0] > 0:
            return True
        # If it's not being played check if game is over and if it was over for the previous event
        elif self.is_game_over() and not self.is_game_over(prev=True):
            return True
        # Check if it's in intermission and the if it was for the previous event
        elif self.is_intermission() and not self.is_intermission(prev=True):
            return True
        else:
            return False

    def time_until_game(self):
        """
        Return the seconds until the game starts

        :return: seconds until game 
        """
        delta = self.start_time - datetime.datetime.utcnow()
        if delta.days >= 0:
            return delta.seconds
        else:
            return 0

    def is_game_over(self, prev=False):
        """
        Check if the game is over for both the html and json pbp. If prev=True check for the previous event
        
        :param prev: Check the game status for the previous event
        
        :return: Boolean - True if over
        """
        if not prev:
            return True if self.html_game_status == self.api_game_status == "Final" else False
        else:
            return True if self.prev_html_game_status == self.prev_api_game_status == "Final" else False

    def is_intermission(self, prev=False):
        """
        Check if in intermission for both the html and json pbp. If prev=True check for the previous event
        
        :param prev: Check the game status for the previous event

        :return: Boolean - True if yes
        """
        if not prev:
            return True if self.html_game_status == self.api_game_status == "Intermission" else False
        else:
            return True if self.prev_html_game_status == self.prev_api_game_status == "Intermission" else False

    def get_pbp(self):
        """
        Return the pbp ensure it's not None
        
        :return: DataFrame
        """
        if isinstance(self.pbp_df, pd.DataFrame):
            return self.pbp_df
        else:
            return pd.DataFrame()

    def get_shifts(self):
        """
        Return the shifts ensure it's not None

        :return: DataFrame
        """
        if isinstance(self.shifts_df, pd.DataFrame):
            return self.shifts_df
        else:
            return pd.DataFrame()


class ScrapeLiveGames:
    """
    Class than contains the info for all the games on a specific day
    
    :param str date: Date of games (ex: 2018-10-30)
    :param bool preseason: If you want to scrape preseason games
    :param bool if_scrape_shifts: Whether or not you want to scrape shifts
    :param list live_games: List of LiveGame objects for 
    :param int pause: Amount to pause after each scraping call
    """

    def __init__(self, date, preseason=False, if_scrape_shifts=False, pause=15, game_ids=list()):
        """
        Initialize the ScrapeLiveGames object with games for the day
        
        :param date: Date 
        :param preseason: If scrape preseason
        :param if_scrape_shifts: Whether to scrape the shifts
        :param pause: time to pause
        :param game_ids: If only want specific games
        """
        # First check date
        check_date_format(date)

        self.user_game_ids = game_ids
        self.date = date
        self.preseason = preseason
        self.if_scrape_shifts = if_scrape_shifts
        self.live_games = self.get_games()          # Hold list of LiveGame objects for that day
        self.pause = pause

    def get_games(self):
        """
        Get initial game info -> Called with object creation. Includes: players, espn_ids, standard game info
        
        :return: Dict - LiveGame objects for all games today
        """
        game_objs = []

        # Get the initial schedule & espn game ids just in case
        games = json_schedule.scrape_schedule(self.date, self.date, live=True, preseason=self.preseason)
        games = self.get_espn_ids(games)

        # Only keep the games we want if the user specified games
        if self.user_game_ids:
            games = [game for game in games if game['game_id'] in self.user_game_ids]

        # Get rosters for each game
        for game in games:
            game_objs.append(LiveGame(game['game_id'], game['start_time'], game['home_team'], game['away_team'],
                                      game['status'], game['espn_id'], self.date, self.if_scrape_shifts))

        return game_objs

    def get_espn_ids(self, games):
        """
        Get espn game ids for all games that day

        :param list games: games today

        :return: Games with corresponding espn game ids
        """
        # Get all espn info
        response = espn_pbp.get_espn_date(self.date)
        game_ids = espn_pbp.get_game_ids(response)
        espn_games = espn_pbp.get_teams(response)

        # Match up
        for i in range(len(games)):
            for j in range(len(espn_games)):
                if games[i]['home_team'] in espn_games[j] or games[i]['away_team'] in espn_games[j]:
                    games[i]['espn_id'] = game_ids[j]

        return games

    def update_live_games(self, force=False, sleep_next=False):
        """
        Scrape the pbp & shifts of ongoing games
        
        :param bool force: Whether or not to force it to scrape even if it's in intermission
        :param bool sleep_next: Sleep until the next game starts
        
        :return: None
        """
        # Check if we need to sleep
        if sleep_next:
            self.sleep_next_game()

        for game in self.live_games:
            game.scrape(force=force)

        time.sleep(self.pause)

    def sleep_next_game(self):
        """
        Sleep until the next game starts. Otherwise just looping and doing nothing
        
        :return: None
        """
        # Get rid of final games...we are looking at current or upcoming games
        non_final_games = [game for game in self.live_games if not game.is_game_over()]

        # Lets get all the games NOT ongoing but aren't over (so scheduled games)
        scheduled_games = [game for game in non_final_games if game.time_until_game() > 0]

        # If all the non-final games haven't started yet let's find the next game
        # Get earliest in the bunch
        if len(scheduled_games) == len(non_final_games):
            min_game = min(scheduled_games, key=lambda x: x.time_until_game())

            if min_game.time_until_game() > 0:
                print("\nSleeping for {} seconds until the next earliest game starts.".format(min_game.time_until_game()))
                time.sleep(min_game.time_until_game())

    def finished(self):
        """
        Check if done with all games
        
        :return: Boolean
        """
        # Count finished games
        finished_games = 0
        for game in self.live_games:
            if game.is_game_over():
                finished_games += 1

        # If the # of finished games == # of total games
        return True if len(self.live_games) == finished_games else False




"""
This module contains functions to scrape the Html game roster for any given game
"""

from bs4 import BeautifulSoup
import hockey_scraper.shared as shared


def get_roster(game_id):
    """
    Given a game_id it returns the raw html
    Ex: http://www.nhl.com/scores/htmlreports/20162017/RO020475.HTM
    
    :param game_id: the game
    
    :return: raw html of game
    """
    game_id = str(game_id)

    page_info = {
        "url": 'http://www.nhl.com/scores/htmlreports/{}{}/RO{}.HTM'.format(game_id[:4], int(game_id[:4]) + 1, game_id[4:]),
        "name": game_id,
        "type": "html_roster",
        "season": game_id[:4],
    }

    return shared.get_file(page_info)


def get_content(roster):
    """
    Uses Beautiful soup to parses the html document.
    Some parsers work for some pages but don't work for others....I'm not sure why so I just try them all here in order
    
    :param roster: doc
    
    :return: players and coaches
    """
    soup = BeautifulSoup(roster, "lxml")
    players = get_players(soup)
    head_coaches = get_coaches(soup)

    if len(players) == 0:
        soup = BeautifulSoup(roster.text, "html.parser")
        players = get_players(soup)
        head_coaches = get_coaches(soup)

        if len(players) == 0:
            soup = BeautifulSoup(roster.text, "html5lib")
            players = get_players(soup)
            head_coaches = get_coaches(soup)

    return players, head_coaches


def fix_name(player):
    """
    Get rid of (A) or (C) when a player has it attached to their name
    
    :param player: list of player info -> [number, position, name]
    
    :return: fixed list
    """
    if player[2].find('(A)') != -1:
        player[2] = player[2][:player[2].find('(A)')].strip()
    elif player[2].find('(C)') != -1:
        player[2] = player[2][:player[2].find('(C)')].strip()

    return player


def get_coaches(soup):
    """
    scrape head coaches
    
    :param soup: html
    
    :return: dict of coaches for game
    """
    coaches = soup.find_all('tr', {'id': "HeadCoaches"})

    # If it picks up nothing just return the empty list
    if not coaches:
        return coaches

    coaches = coaches[0].find_all('td')

    return {
        'Away': coaches[1].get_text(),
        'Home': coaches[3].get_text()
    }


def get_players(soup):
    """
    scrape roster for players 
    
    :param soup: html
    
    :return: dict for home and away players
    """
    tables = soup.findAll('table', {'align': 'center', 'border': '0', 'cellpadding': '0', 'cellspacing': '0', 'width': '100%'})

    # If it picks up nothing just return the empty list
    if not tables:
        return tables

    """
    There are 5 tables which correspond to the above criteria.
    tables[0] is nothing
    tables[1] is away starters
    tables[2] is home starters
    tables[3] is away scratches
    tables[4] is home scratches
    """

    del tables[0]
    player_info = [table.find_all('td') for table in tables]

    player_info = [[x.get_text() for x in group] for group in player_info]

    # Make list of list of 3 each. The three are: number, position, name (in that order)
    player_info = [[group[i:i+3] for i in range(0, len(group), 3)] for group in player_info]

    # Get rid of header column
    player_info = [[player for player in group if player[0] != '#'] for group in player_info]

    # Add whether the player was a scratch
    # 2 and 3 hold the scratches
    for i in range(len(player_info)):
        for j in range(len(player_info[i])):
            if i == 2 or i == 3:
                player_info[i][j].append(True)
            else:
                player_info[i][j].append(False)

    players = {'Away': player_info[0], 'Home': player_info[1]}

    # Scratches aren't always included
    if len(player_info) == 4:
        players['Away'] += player_info[2]
        players['Home'] += player_info[3]

    # For those with (A) or (C) in name field get rid of it
    # First condition is to control when we get whitespace as one of the indices
    players['Away'] = [fix_name(i) if i[0] != u'\xa0' else i for i in players['Away']]
    players['Home'] = [fix_name(i) if i[0] != u'\xa0' else i for i in players['Home']]

    # Get rid when just whitespace
    players['Away'] = [i for i in players['Away'] if i[0] != u'\xa0']
    players['Home'] = [i for i in players['Home'] if i[0] != u'\xa0']

    return players


def scrape_roster(game_id):
    """
    For a given game scrapes the roster
    
    :param game_id: id for game
    
    :return: dict of players (home and away) an dict for both head coaches 
    """
    roster = get_roster(game_id)

    if not roster:
        shared.print_warning("Roster for game {} is either not there or can't be obtained".format(game_id))
        return None

    try:
        players, head_coaches = get_content(roster)
    except Exception as e:
        shared.print_warning('Error parsing Roster for game {} {}'.format(game_id, e))
        return None

    return {'players': players, 'head_coaches': head_coaches}



"""
This module contains functions to scrape the Html Play by Play for any given game
"""

import pandas as pd
from bs4 import BeautifulSoup, SoupStrainer
import re
import hockey_scraper.shared as shared


def cur_game_status(doc):
    """
    Return the game status
    
    :param doc: Html text
    
    :return: String -> one of ['Final', 'Intermission', 'Progress']
    """
    soup = BeautifulSoup(doc, "lxml")
    tables = soup.find_all('table', {'id': "GameInfo"})
    tds = tables[0].find_all('td')
    status = tds[-1].text

    # 'End' - in there means an Intermission
    # 'Final' - Game is over
    # Otherwise - It's either in progress or b4 the game started
    if 'end' in status.lower():
        return 'Intermission'
    elif 'final' in status.lower():
        return 'Final'
    else:
        return 'Live'


def get_pbp(game_id):
    """
    Given a game_id it returns the raw html
    Ex: http://www.nhl.com/scores/htmlreports/20162017/PL020475.HTM
    
    :param game_id: the game
    
    :return: raw html of game
    """
    game_id = str(game_id)
    url = 'http://www.nhl.com/scores/htmlreports/{}{}/PL{}.HTM'.format(game_id[:4], int(game_id[:4]) + 1, game_id[4:])

    page_info = {
        "url": url,
        "name": game_id,
        "type": "html_pbp",
        "season": game_id[:4],
    }

    return shared.get_file(page_info)


def get_soup(game_html):
    """
    Uses Beautiful soup to parses the html document.
    Some parsers work for some pages but don't work for others....I'm not sure why so I just try them all here in order
    
    :param game_html: html doc
    
    :return: "soupified" html 
    """
    strainer = SoupStrainer('td', attrs={'class': re.compile(r'bborder')})
    soup = BeautifulSoup(game_html, "lxml", parse_only=strainer)
    soup = soup.select('td.+.bborder')

    if len(soup) == 0:
        soup = BeautifulSoup(game_html, "html.parser", parse_only=strainer)
        soup = soup.select('td.+.bborder')

        if len(soup) == 0:
            soup = BeautifulSoup(game_html, "html5lib")
            soup = soup.select('td.+.bborder')

    return soup


def strip_html_pbp(td):
    """
    Strip html tags and such 
    
    :param td: pbp
    
    :return: list of plays (which contain a list of info) stripped of html
    """

    for y in range(len(td)):
        # Get the 'br' tag for the time column...this get's us time remaining instead of elapsed and remaining combined
        if y == 3:
            td[y] = td[y].get_text()   # This gets us elapsed and remaining combined-< 3:0017:00
            index = td[y].find(':')
            td[y] = td[y][:index+3]
        elif (y == 6 or y == 7) and td[0] != '#':
            # 6 & 7-> These are the player 1 ice one's
            # The second statement controls for when it's just a header
            baz = td[y].find_all('td')
            bar = [baz[z] for z in range(len(baz)) if z % 4 != 0]  # Because of previous step we get repeats...delete some

            # The setup in the list is now: Name/Number->Position->Blank...and repeat
            # Now strip all the html
            players = []
            for i in range(len(bar)):
                if i % 3 == 0:
                    try:
                        name = return_name_html(bar[i].find('font')['title'])
                        number = bar[i].get_text().strip('\n')  # Get number and strip leading/trailing newlines
                    except KeyError:
                        name = ''
                        number = ''
                elif i % 3 == 1:
                    if name != '':
                        position = bar[i].get_text()
                        players.append([name, number, position])

            td[y] = players
        else:
            td[y] = td[y].get_text()

    return td


def clean_html_pbp(html):
    """
    Get rid of html and format the data
    
    :param html: the requested url
    
    :return: a list with all the info
    """
    soup = get_soup(html)

    # Create a list of lists (each length 8)...corresponds to 8 columns in html pbp
    td = [soup[i:i + 8] for i in range(0, len(soup), 8)]

    cleaned_html = [strip_html_pbp(x) for x in td]

    return cleaned_html


def add_home_zone(event_dict, home_team):
    """
    Determines the zone relative to the home team and add it to event.
    
    Keep in mind that the 'ev_zone' recorded is the zone relative to the event team. And for blocks the NHL counts
    the ev_team as the blocking team (I like counting the shooting team for blocks). Therefore, when it's the home team
    the zone only gets flipped when it's a block. For away teams it's the opposite. Think about it...
    
    :param event_dict: dict of event info
    :param home_team: home team
    
    :return: None
    """
    ev_team = event_dict['Ev_Team']
    ev_zone = event_dict['Ev_Zone']
    event = event_dict['Event']

    # Return if we got nothing in there
    # Also just make the home zone nothing too
    if ev_zone == '':
        event_dict['Home_Zone'] = ''
        return

    # When it's either: The away team and not a block or the home team and a block
    if (ev_team != home_team and event != 'BLOCK') or (ev_team == home_team and event == 'BLOCK'):
        if ev_zone == 'Off':
            event_dict['Home_Zone'] = 'Def'
        elif ev_zone == 'Def':
            event_dict['Home_Zone'] = 'Off'
        else:
            event_dict['Home_Zone'] = ev_zone
    else:
        event_dict['Home_Zone'] = ev_zone


def add_zone(event_dict, play_description):
    """
    Determine which zone the play occurred in (unless one isn't listed) and add it to dict
    
    :param event_dict: dict of event info
    :param play_description: the zone would be included here
    
    :return: Off, Def, Neu, or NA
    """
    s = [x.strip() for x in play_description.split(',')]  # Split by comma's into a list
    zone = [x for x in s if 'Zone' in x]  # Find if list contains which zone

    if not zone:
        event_dict['Ev_Zone'] = ''
        return

    if zone[0].find("Off") != -1:
        event_dict['Ev_Zone'] = 'Off'
    elif zone[0].find("Neu") != -1:
        event_dict['Ev_Zone'] = 'Neu'
    elif zone[0].find("Def") != -1:
        event_dict['Ev_Zone'] = 'Def'


def add_type(event_dict, event, players, home_team):
    """
    Add "type" for event -> either a penalty or a shot type
    
    :param event_dict: dict of event info
    :param event: list with parsed event info
    :param players: dict of home and away players in game
    :param home_team: home team for game
    
    :return: None
    """
    if 'PENL' in event[4]:
        event_dict['Type'] = get_penalty(event[5], players, home_team)
    else:
        event_dict['Type'] = shot_type(event[5]).upper()


def add_strength(event_dict, home_players, away_players):
    """
    Get strength for event -> It's home then away
    
    :param event_dict: dict of event info
    :param home_players: list of players for home team
    :param away_players: list of players for away team
    
    :return: None
    """
    try:
        home_skaters = event_dict['Home_Players'] - 1 if event_dict['Home_Goalie'] != '' else len(home_players)
        away_skaters = event_dict['Away_Players'] - 1 if event_dict['Away_Goalie'] != '' else len(away_players)
    except KeyError:
        # Getting a key error here means that home/away goalie isn't there...which means home/away players are empty
        home_skaters = 0
        away_skaters = 0

    event_dict['Strength'] = 'x'.join([str(home_skaters), str(away_skaters)])


def add_event_team(event_dict, event):
    """
    Add event team for event 
    
    :param event_dict: dict of event info
    :param event: list with parsed event info
    
    :return: None
    """
    if event_dict['Event'] in ['GOAL', 'SHOT', 'MISS', 'BLOCK', 'PENL', 'FAC', 'HIT', 'TAKE', 'GIVE']:
        # Split the description and take the first thing (which is the team)
        event_dict['Ev_Team'] = event[5].split()[0]
    else:
        event_dict['Ev_Team'] = ''


def add_period(event_dict, event):
    """
    Add period for event 
    
    :param event_dict: dict of event info
    :param event: list with parsed event info
    
    :return: None
    """
    try:
        event_dict['Period'] = int(event[1])
    except ValueError:
        event_dict['Period'] = 0


def add_time(event_dict, event):
    """
    Fill in time and seconds elapsed
    
    :param event_dict: dict of parsed event stuff
    :param event: event info from pbp
    
    :return: None
    """
    event_dict['Time_Elapsed'] = str(event[3])

    if event[3] != '':
        event_dict['Seconds_Elapsed'] = shared.convert_to_seconds(event[3])
    else:
        event_dict['Seconds_Elapsed'] = ''


def add_score(event_dict, event, current_score, home_team):
    """
    Change if someone scored...also change current score
    
    :param event_dict: dict of parsed event stuff
    :param event: event info from pbp
    :param current_score: current score in game
    :param home_team: home team for game
    
    :return: None
    """

    # If it's a goal change the score
    if event[4] == 'GOAL':
        if event_dict['Ev_Team'] == home_team:
            current_score['Home'] += 1
        else:
            current_score['Away'] += 1

    event_dict['Home_Score'] = current_score['Home']
    event_dict['Away_Score'] = current_score['Away']
    event_dict['score_diff'] = current_score['Home'] - current_score['Away']


def get_penalty(play_description, players, home_team):
    """
    Get the penalty info
    
    :param play_description: description of play field
    :param players: all players with info
    :param home_team: home team for game
    
    :return: penalty info
    """
    # First check if it's a bench
    if "bench" in play_description or "TEAM" in play_description:
        beg_penalty_index = play_description.find("TEAM") + 5
        return play_description[beg_penalty_index: play_description.find(')') + 1]
    else:
        # If it's not a bench penl we look for the player who took the penalty
        # Get Number, and name for player who took the penalty
        num_regex = re.compile(r'#(\d+)')
        numbers = num_regex.findall(play_description)

        # If they don't have any players listed, then the description if fucked up and we got nothing
        if not numbers:
            return ''
        else:
            player = get_player_name(numbers[0], players, play_description[:3], home_team)

        # Check if the number and player match up
        if play_description.find(player['last_name']) != -1:
            # beg_penalty_index is right after the penalty taker's last name (+1 for whitespace)
            # Then we take from after his last name to right after the parentheses
            beg_penalty_index = play_description.find(player['last_name']) + len(player['last_name']) + 1
            return play_description[beg_penalty_index: play_description.find(')')+1]
        else:
            # This uses my old method...it falls apart for players like "Del Zotto"
            pen_regex = re.compile(r'.{3}\s+#\d+\s+\w+\s+(.*)\)')
            penalty = pen_regex.findall(play_description)
            return penalty[0] + ')' if penalty else ''


def get_player_name(number, players, team, home_team):
    """
    This function is used for the description field in the html. Given a last name and a number it return the player's 
    full name and id.
    
    :param number: player's number
    :param players: all players with info
    :param team: team of player
    :param home_team: home team
    
    :return: dict with full and and id
    """
    venue = "Home" if team == home_team else "Away"

    # Get the info when we get the same number for that team
    player = [{'name': name, 'id': players[venue][name]['id'], 'last_name': players[venue][name]['last_name']}
              for name in players[venue].keys() if players[venue][name]['number'] == number]

    # Control for when the name can't be found
    if not player:
        player = [{'name': '', 'id': '', 'last_name': ''}]

    return player[0]


def if_valid_event(event):
    """
    Checks if it's a valid event ('#' is meaningless and I don't like those other one's) to parse
    
    Don't remember what 'GOFF' is but 'EGT' is for emergency goaltender. The reason I get rid of it is because it's not
    in the json and there's another 'EGPID' that can be found in both (not sure why 'EGT' exists then).
    
    Events 'PGSTR', 'PGEND', and 'ANTHEM' have been included at the start of each game for the 2017 season...I have no
    idea why. 
     
    :param event: list of stuff in pbp
    
    :return: boolean 
    """
    if event[0] != '#' and event[4] not in ['GOFF', 'EGT', 'PGSTR', 'PGEND', 'ANTHEM']:
        return True
    else:
        return False


def return_name_html(info):
    """
    In the PBP html the name is in a format like: 'Center - MIKE RICHARDS'
    Some also have a hyphen in their last name so can't just split by '-'
    
    :param info: position and name
    
    :return: name
    """
    s = info.index('-')  # Find first hyphen
    return info[s + 1:].strip(' ')  # The name should be after the first hyphen


def shot_type(play_description):
    """
    Determine which zone the play occurred in (unless one isn't listed)
    
    :param play_description: the type would be in here
    
    :return: the type if it's there (otherwise just NA)
    """
    types = ['wrist', 'snap', 'slap', 'deflected', 'tip-in', 'backhand', 'wrap-around']

    play_description = [x.strip() for x in play_description.split(',')]  # Strip leading and trailing whitespace
    play_description = [i.lower() for i in play_description]  # Convert to lowercase

    for p in play_description:
        if p in types:
            if p == 'wrist' or p == 'slap' or p == 'snap':
                return ' '.join([p, 'shot'])
            else:
                return p
    return ''


def parse_fac(description, players, ev_team, home_team):
    """
    Parse the description field for a face-off
    MTL won Neu. Zone - MTL #11 GOMEZ vs TOR #37 BRENT
    
    :param description: Play Description 
    :param players: players in game
    :param ev_team: Event Team
    :param home_team: Home Team for game
    
    :return: Dict with info
    """
    event_info = {}

    regex = re.compile(r'(.{3})\s+#(\d+)')
    desc = regex.findall(description)  # [[Team, num], [Team, num]]

    if ev_team == desc[0][0]:
        p1 = get_player_name(desc[0][1], players, desc[0][0], home_team)
        p2 = get_player_name(desc[1][1], players, desc[1][0], home_team)
    else:
        p1 = get_player_name(desc[1][1], players, desc[1][0], home_team)
        p2 = get_player_name(desc[0][1], players, desc[0][0], home_team)

    event_info['p1_name'] = p1['name']
    event_info['p1_ID'] = p1['id']
    event_info['p2_name'] = p2['name']
    event_info['p2_ID'] = p2['id']

    return event_info


def parse_shot_miss_take_give(description, players, ev_team, home_team):
    """
    Parse the description field for a: SHOT, MISS, TAKE, GIVE
    
    MTL ONGOAL - #81 ELLER, Wrist, Off. Zone, 11 ft.
    ANA #23 BEAUCHEMIN, Slap, Wide of Net, Off. Zone, 42 ft.
    TOR GIVEAWAY - #35 GIGUERE, Def. Zone
    TOR TAKEAWAY - #9 ARMSTRONG, Off. Zone
    
    :param description: Play Description 
    :param players: players in game
    :param ev_team: Event Team
    :param home_team: Home Team for game
    
    :return: Dict with info
    """
    event_info = {}

    regex = re.compile(r'(\d+)')
    desc = regex.search(description).groups()  # num

    p = get_player_name(desc[0], players, ev_team, home_team)
    event_info['p1_name'] = p['name']
    event_info['p1_ID'] = p['id']

    return event_info


def parse_hit(description, players, home_team):
    """
    Parse the description field for a HIT

    MTL #20 O'BYRNE HIT TOR #18 BROWN, Def. Zone

    :param description: Play Description 
    :param players: players in game
    :param home_team: Home Team for game

    :return: Dict with info
    """
    event_info = {}

    regex = re.compile(r'(.{3})\s+#(\d+)')
    desc = regex.findall(description)  # [[Team, num], [Team, num]]

    p1 = get_player_name(desc[0][1], players, desc[0][0], home_team)
    event_info['p1_name'] = p1['name']
    event_info['p1_ID'] = p1['id']

    if len(desc) > 1:
        p2 = get_player_name(desc[1][1], players, desc[1][0], home_team)
        event_info['p2_name'] = p2['name']
        event_info['p2_ID'] = p2['id']

    return event_info


def parse_block(description, players, home_team):
    """
    Parse the description field for a BLOCK
    
    MTL #76 SUBBAN BLOCKED BY TOR #2 SCHENN, Wrist, Def. Zone

    :param description: Play Description 
    :param players: players in game
    :param home_team: Home Team for game

    :return: Dict with info
    """
    event_info = {}

    regex = re.compile(r'(.{3})\s+#(\d+)')
    desc = regex.findall(description)  # [[Team, num], [Team, num]]

    p1 = get_player_name(desc[len(desc) - 1][1], players, desc[len(desc) - 1][0], home_team)
    event_info['p1_name'] = p1['name']
    event_info['p1_ID'] = p1['id']

    if len(desc) > 1:
        p2 = get_player_name(desc[0][1], players, desc[0][0], home_team)
        event_info['p2_name'] = p2['name']
        event_info['p2_ID'] = p2['id']

    return event_info


def parse_goal(description, players, ev_team, home_team):
    """
    Parse the description field for a GOAL
    
    TOR #81 KESSEL(1), Wrist, Off. Zone, 14 ft. Assists: #42 BOZAK(1); #8 KOMISAREK(1)
    
    :param description: Play Description 
    :param players: players in game
    :param ev_team: Event Team
    :param home_team: Home Team for game

    :return: Dict with info
    """
    event_info = {}

    regex = re.compile(r'#(\d+)\s+')
    desc = regex.findall(description)  # [num, ?, ?] -> ranging from 1 to 3 indices

    p1 = get_player_name(desc[0], players, ev_team, home_team)
    event_info['p1_name'] = p1['name']
    event_info['p1_ID'] = p1['id']

    if len(desc) >= 2:
        p2 = get_player_name(desc[1], players, ev_team, home_team)
        event_info['p2_name'] = p2['name']
        event_info['p2_ID'] = p2['id']

        if len(desc) == 3:
            p3 = get_player_name(desc[2], players, ev_team, home_team)
            event_info['p3_name'] = p3['name']
            event_info['p3_ID'] = p3['id']

    return event_info


def parse_penalty(description, players, home_team):
    """
    Parse the description field for a Penalty

    MTL #81 ELLER Hooking(2 min), Def. Zone Drawn By: TOR #11 SJOSTROM

    :param description: Play Description 
    :param players: players in game
    :param home_team: Home Team for game

    :return: Dict with info
    """
    event_info = {}

    # Check if it's a Bench/Team Penalties
    if "bench" in description or "TEAM" in description:
        event_info['p1_name'] = 'Team'
    else:
        # Standard Penalty
        regex = re.compile(r'(.{3})\s+#(\d+)')
        desc = regex.findall(description)  # [[team, num], ?[team, num]] -> Either one to three indices

        if desc:
            p1 = get_player_name(desc[0][1], players, desc[0][0], home_team)
            event_info['p1_name'] = p1['name']
            event_info['p1_ID'] = p1['id']

            # When there are three the penalty was served by someone else
            # The Person who served the penalty is placed as the 3rd event player
            if len(desc) == 3:
                p3 = get_player_name(desc[1][1], players, desc[0][0], home_team)
                event_info['p3_name'] = p3['name']
                event_info['p3_ID'] = p3['id']

                p2 = get_player_name(desc[2][1], players, desc[2][0], home_team)
                event_info['p2_name'] = p2['name']
                event_info['p2_ID'] = p2['id']
            elif len(desc) == 2:
                p2 = get_player_name(desc[1][1], players, desc[1][0], home_team)
                event_info['p2_name'] = p2['name']
                event_info['p2_ID'] = p2['id']

    return event_info


def add_event_players(event_dict, event, players, home_team):
    """
    Add players involved in the event to event_dict
    
    :param event_dict: dict of parsed event stuff
    :param event: fixed up html
    :param players: dict of players and id's
    :param home_team: home team
    
    :return: None
    """
    description = event[5].strip()
    ev_team = description.split()[0]
    event_info = {}

    if event[4] == 'FAC':
        event_info = parse_fac(description, players, ev_team, home_team)
    elif event[4] in ['SHOT', 'MISS', 'GIVE', 'TAKE']:
        event_info = parse_shot_miss_take_give(description, players, ev_team, home_team)
    elif event[4] == 'HIT':
        event_info = parse_hit(description, players, home_team)
    elif event[4] == 'BLOCK':
        event_info = parse_block(description, players, home_team)
    elif event[4] == 'GOAL':
        event_info = parse_goal(description, players, ev_team, home_team)
    elif event[4] == 'PENL':
        event_info = parse_penalty(description, players, home_team)

    # Transfer info over
    for key in event_info.keys():
        event_dict[key] = event_info[key]


def populate_players(event_dict, players, away_players, home_players):
    """
    Populate away and home player info (and num skaters on each side)
    
    :param event_dict: dict with event info
    :param players: all players in game and info
    :param away_players: players for away team
    :param home_players: players for home team
    
    :return: None
    """
    for venue in ['home', 'away']:
        for j in range(6):
            try:
                if venue == "home":
                    name = shared.fix_name(home_players[j][0].upper())
                    foo = 'Home'
                else:
                    name = shared.fix_name(away_players[j][0].upper())
                    foo = 'Away'

                event_dict['{}Player{}'.format(venue, j + 1)] = name
                event_dict['{}Player{}_id'.format(venue, j + 1)] = players[foo][name]['id']
            except KeyError:
                event_dict['{}Player{}_id'.format(venue, j + 1)] = 'NA'
            except IndexError:
                event_dict['{}Player{}'.format(venue, j + 1)] = ''
                event_dict['{}Player{}_id'.format(venue, j + 1)] = ''

    # Did this because above method assumes the goalie is at end of player list
    for x in away_players:
        if x[2] == 'G':
            event_dict['Away_Goalie'] = shared.fix_name(x[0].upper())
            try:
                event_dict['Away_Goalie_Id'] = players['Away'][event_dict['Away_Goalie']]['id']
            except KeyError:
                event_dict['Away_Goalie_Id'] = 'NA'
        else:
            event_dict['Away_Goalie'] = ''
            event_dict['Away_Goalie_Id'] = ''

    for x in home_players:
        if x[2] == 'G':
            event_dict['Home_Goalie'] = shared.fix_name(x[0].upper())
            try:
                event_dict['Home_Goalie_Id'] = players['Home'][event_dict['Home_Goalie']]['id']
            except KeyError:
                event_dict['Home_Goalie_Id'] = 'NA'
        else:
            event_dict['Home_Goalie'] = ''
            event_dict['Home_Goalie_Id'] = ''

    event_dict['Away_Players'] = len(away_players)
    event_dict['Home_Players'] = len(home_players)


def parse_event(event, players, home_team, current_score):
    """
    Receives an event and parses it
    
    :param event: event type
    :param players: players in game
    :param home_team: home team
    :param current_score: current score for both teams
    
    :return: dict with info
    """
    event_dict = dict()

    away_players = event[6]
    home_players = event[7]

    event_dict['Description'] = event[5]
    event_dict['Event'] = str(event[4])

    add_period(event_dict, event)
    add_time(event_dict, event)
    add_event_team(event_dict, event)
    add_score(event_dict, event, current_score, home_team)
    populate_players(event_dict, players, away_players, home_players)
    add_strength(event_dict, home_players, away_players)
    add_type(event_dict, event, players, home_team)
    add_zone(event_dict, event[5])
    add_home_zone(event_dict, home_team)

    # Sometimes it's empty...(they seem to sometimes/always have a whitespace char)
    if len(event_dict['Description']) > 1:
        add_event_players(event_dict, event, players, home_team)

    return event_dict


def parse_html(html, players, teams):
    """
    Parse html game pbp
    
    :param html: raw html
    :param players: players in the game (from json pbp)
    :param teams: dict with home and away teams
    
    :return: DataFrame with info
    """
    columns = ['Period', 'Event', 'Description', 'Time_Elapsed', 'Seconds_Elapsed', 'Strength', 'Ev_Zone', 'Type',
               'Ev_Team', 'Home_Zone', 'Away_Team', 'Home_Team', 'p1_name', 'p1_ID', 'p2_name', 'p2_ID', 'p3_name',
               'p3_ID', 'awayPlayer1', 'awayPlayer1_id', 'awayPlayer2', 'awayPlayer2_id', 'awayPlayer3', 'awayPlayer3_id',
               'awayPlayer4', 'awayPlayer4_id', 'awayPlayer5', 'awayPlayer5_id', 'awayPlayer6', 'awayPlayer6_id',
               'homePlayer1', 'homePlayer1_id', 'homePlayer2', 'homePlayer2_id', 'homePlayer3', 'homePlayer3_id',
               'homePlayer4', 'homePlayer4_id', 'homePlayer5', 'homePlayer5_id', 'homePlayer6', 'homePlayer6_id',
               'Away_Goalie', 'Away_Goalie_Id', 'Home_Goalie', 'Home_Goalie_Id', 'Away_Players', 'Home_Players',
               'Away_Score', 'Home_Score']

    current_score = {'Home': 0, 'Away': 0}
    events = [parse_event(event, players, teams['Home'], current_score) for event in html if if_valid_event(event)]

    df = pd.DataFrame(list(events), columns=columns)

    # This is seen sometimes...it's a duplicate row
    df.drop(df[df.Time_Elapsed == '-16:0-'].index, inplace=True)

    df['Away_Team'] = teams['Away']
    df['Home_Team'] = teams['Home']

    return df


def scrape_pbp(game_html, game_id, players, teams):
    """
    Scrape the data for the pbp

    :param game_html: Html doc for the game
    :param game_id: game to scrape
    :param players: dict with player info
    :param teams: dict with home and away teams

    :return: DataFrame of game info or None if it fails
    """
    if not game_html:
        shared.print_warning("Html pbp for game {} is either not there or can't be obtained".format(game_id))
        return None

    cleaned_html = clean_html_pbp(game_html)
    if len(cleaned_html) == 0:
        shared.print_warning("Html pbp contains no plays, this game can't be scraped")
        return None

    try:
        game_df = parse_html(cleaned_html, players, teams)
    except Exception as e:
        shared.print_warning('Error parsing Html pbp for game {} {}'.format(game_id, e))
        return None

    return game_df


def scrape_game_live(game_id, players, teams):
    """
    Scrape the data for the game when it's live
    
    :param game_id: game to scrape
    :param players: dict with player info
    :param teams: dict with home and away teams
    
    :return: Tuple - get_pbp(), cur_game_status()
    """
    game_html = get_pbp(game_id)
    return scrape_pbp(game_html, game_id, players, teams), cur_game_status(game_html)


def scrape_game(game_id, players, teams):
    """ 
    Scrape the data for the game when not live
    
    :param game_id: game to scrape
    :param players: dict with player info
    :param teams: dict with home and away teams
    
    :return: DataFrame of game info or None if it fails
    """
    game_html = get_pbp(game_id)
    return scrape_pbp(game_html, game_id, players, teams)







"""
This module contains functions to scrape the Html Toi Tables (or shifts) for any given game
"""

import pandas as pd
from bs4 import BeautifulSoup
import re
import hockey_scraper.shared as shared


def get_shifts(game_id):
    """
    Given a game_id it returns a the shifts for both teams
    Ex: http://www.nhl.com/scores/htmlreports/20162017/TV020971.HTM
    
    :param game_id: the game
    
    :return: Shifts or None
    """
    game_id = str(game_id)
    home_url = 'http://www.nhl.com/scores/htmlreports/{}{}/TH{}.HTM'.format(game_id[:4], int(game_id[:4])+1, game_id[4:])
    away_url = 'http://www.nhl.com/scores/htmlreports/{}{}/TV{}.HTM'.format(game_id[:4], int(game_id[:4])+1, game_id[4:])

    page_info = {
        "url": home_url,
        "name": game_id,
        "type": "html_shifts_home",
        "season": game_id[:4],
    }

    # Get info for home shifts
    home = shared.get_file(page_info)

    # Change info for scraping away page and scrape it
    page_info["type"] = "html_shifts_away"
    page_info["url"] = away_url
    away = shared.get_file(page_info)

    return home, away


def get_soup(shifts_html):
    """
    Uses Beautiful soup to parses the html document.
    Some parsers work for some pages but don't work for others....I'm not sure why so I just try them all here in order
    
    :param shifts_html: html doc
    
    :return: "soupified" html and player_shifts portion of html (it's a bunch of td tags)
    """
    soup = BeautifulSoup(shifts_html, "lxml")
    td = soup.findAll(True, {'class': ['playerHeading + border', 'lborder + bborder']})

    if len(td) == 0:
        soup = BeautifulSoup(shifts_html, "html.parser")
        td = soup.findAll(True, {'class': ['playerHeading + border', 'lborder + bborder']})

        if len(td) == 0:
            soup = BeautifulSoup(shifts_html, "html5lib")
            td = soup.findAll(True, {'class': ['playerHeading + border', 'lborder + bborder']})

    return td, get_teams(soup)


def get_teams(soup):
    """
    Return the team for the TOI tables and the home team
    
    :param soup: souped up html
    
    :return: list with team and home team
    """
    team = soup.find('td', class_='teamHeading + border')  # Team for shifts
    team = team.get_text()

    # Get Home Team
    teams = soup.find_all('td', {'align': 'center', 'style': 'font-size: 10px;font-weight:bold'})
    regex = re.compile(r'>(.*)<br/?>')
    home_team = regex.findall(str(teams[7]))

    return [team, home_team[0]]


def analyze_shifts(shift, name, team, home_team, player_ids):
    """
    Analyze shifts for each player when using.
    Prior to this each player (in a dictionary) has a list with each entry being a shift.

    :param shift: info on shift
    :param name: player name
    :param team: given team
    :param home_team: home team for given game
    :param player_ids: dict with info on players
    
    :return: dict with info for shift
    """
    shifts = dict()

    regex = re.compile('\d+')  # Used to check if something contains a number

    shifts['Player'] = name.upper()
    shifts['Period'] = '4' if shift[1] == 'OT' else shift[1]
    shifts['Team'] = shared.get_team(team.strip(' '))
    shifts['Start'] = shared.convert_to_seconds(shift[2].split('/')[0])
    shifts['Duration'] = shared.convert_to_seconds(shift[4].split('/')[0])

    # I've had problems with this one...if there are no digits the time is fucked up
    if regex.findall(shift[3].split('/')[0]):
        shifts['End'] = shared.convert_to_seconds(shift[3].split('/')[0])
    else:
        shifts['End'] = shifts['Start'] + shifts['Duration']

    try:
        if home_team == team:
            shifts['Player_Id'] = player_ids['Home'][name.upper()]['id']
        else:
            shifts['Player_Id'] = player_ids['Away'][name.upper()]['id']
    except KeyError:
        shifts['Player_Id'] = ''

    return shifts


def parse_html(html, player_ids, game_id):
    """
    Parse the html
    
    Note: Don't fuck with this!!! I'm not exactly sure how or why but it works. 
    
    :param html: cleaned up html
    :param player_ids: dict of home and away players
    :param game_id: id for game
    
    :return: DataFrame with info
    """
    columns = ['Game_Id', 'Player', 'Player_Id', 'Period', 'Team', 'Start', 'End', 'Duration']
    df = pd.DataFrame(columns=columns)

    td, teams = get_soup(html)

    team = teams[0]
    home_team = teams[1]
    players = dict()

    # The list 'td' is laid out with player name followed by every component of each shift. Each shift contains:
    # shift #, Period, begin, end, and duration. The shift event isn't included.
    for t in td:
        t = t.get_text()
        if ',' in t:     # If it has a comma in it we know it's a player's name...so add player to dict
            name = t
            # Just format the name normally...it's coded as: 'num last_name, first_name'
            name = name.split(',')
            name = ' '.join([name[1].strip(' '), name[0][2:].strip(' ')])
            name = shared.fix_name(name)
            players[name] = dict()
            players[name]['number'] = name[0][:2].strip()
            players[name]['Shifts'] = []
        else:
            # Here we add all the shifts to whatever player we are up to
            players[name]['Shifts'].extend([t])

    for key in players.keys():
        # Create a list of lists (each length 5)...corresponds to 5 columns in html shifts
        players[key]['Shifts'] = [players[key]['Shifts'][i:i + 5] for i in range(0, len(players[key]['Shifts']), 5)]

        # Parse each shift
        shifts = [analyze_shifts(shift, key, team, home_team, player_ids) for shift in players[key]['Shifts']]
        df = df.append(shifts, ignore_index=True)

    df['Game_Id'] = str(game_id)[5:]
    return df


def scrape_game(game_id, players):
    """
    Scrape the game. 
    
    :param game_id: id for game
    :param players: list of players
    
    :return: DataFrame with info for the game
    """
    columns = ['Game_Id', 'Period', 'Team', 'Player', 'Player_Id', 'Start', 'End', 'Duration']

    home_html, away_html = get_shifts(game_id)

    if home_html is None or away_html is None:
        shared.print_warning("Html shifts for game {} is either not there or can't be obtained".format(game_id))
        return None

    try:
        away_df = parse_html(away_html, players, game_id)
        home_df = parse_html(home_html, players, game_id)
    except Exception as e:
        shared.print_warning('Error parsing Html shifts for game {} {}'.format(game_id, e))
        return None

    # Combine the two
    game_df = pd.concat([away_df, home_df], ignore_index=True)
    game_df = pd.DataFrame(game_df, columns=columns)

    game_df = game_df.sort_values(by=['Period', 'Start'], ascending=[True, True])
    game_df = game_df.reset_index(drop=True)

    return game_df
